use crate::vm::dm_type::DmType;
use crate::vm::object::DvmObject;
use std::fmt::Debug;
use std::rc::Rc;

#[derive(Debug, Eq)]
pub struct DmFn {
    fqn: String,
    short_name: String,
    byte_code: Vec<u8>,
    number_used_registers: usize,
    return_register: Option<u8>,
}

impl DmFn {
    pub fn new(
        fqn: &str,
        short_name: &str,
        byte_code: Vec<u8>,
        number_used_registers: usize,
        return_register: Option<u8>,
    ) -> Self {
        DmFn {
            fqn: fqn.to_string(),
            short_name: short_name.to_string(),
            byte_code,
            number_used_registers,
            return_register,
        }
    }

    pub fn fqn(&self) -> &str {
        self.fqn.as_str()
    }

    pub fn short_name(&self) -> &str {
        self.short_name.as_str()
    }

    pub fn byte_code(&self) -> &Vec<u8> {
        &self.byte_code
    }

    pub fn number_used_registers(&self) -> usize {
        self.number_used_registers
    }

    pub fn return_register(&self) -> Option<u8> {
        self.return_register
    }
}

impl PartialEq for DmFn {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

#[derive(Debug, Eq)]
pub struct DmMethod {
    dm_fn: Rc<DmFn>,
    implements: Option<Rc<DmVirtualMethod>>,
}

impl PartialEq for DmMethod {
    fn eq(&self, other: &Self) -> bool {
        self.dm_fn == other.dm_fn
    }
}

impl DmMethod {
    pub fn new(dm_fn: DmFn, implements: Option<Rc<DmVirtualMethod>>) -> Self {
        DmMethod {
            dm_fn: Rc::new(dm_fn),
            implements,
        }
    }

    pub fn dm_fn(&self) -> Rc<DmFn> {
        self.dm_fn.clone()
    }
}

#[derive(Debug, Eq)]
pub struct DmInterface {
    fqn: String,
    short_name: String,
    functions: Vec<Rc<DmFn>>,
    virtual_methods: Vec<DmVirtualMethod>,
}

impl PartialEq for DmInterface {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmInterface {
    pub fn new(fqn: &str, short_name: &str) -> Self {
        DmInterface {
            fqn: fqn.to_string(),
            short_name: short_name.to_string(),
            functions: Vec::new(),
            virtual_methods: Vec::new(),
        }
    }

    pub fn fqn(&self) -> &str {
        self.fqn.as_str()
    }

    pub fn short_name(&self) -> &str {
        self.short_name.as_str()
    }

    pub fn add_function(&mut self, dm_fn: DmFn) {
        self.functions.push(Rc::new(dm_fn));
    }

    pub fn functions(&self) -> &Vec<Rc<DmFn>> {
        &self.functions
    }

    pub fn add_method(&mut self, dm_virtual_method: DmVirtualMethod) {
        self.virtual_methods.push(dm_virtual_method);
    }

    pub fn virtual_methods(&self) -> &Vec<DmVirtualMethod> {
        &self.virtual_methods
    }
}

#[derive(Debug, Eq)]
pub struct DmVirtualMethod {
    fqn: String,
    short_name: String,
}

impl PartialEq for DmVirtualMethod {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmVirtualMethod {
    pub fn new(fqn: &str, short_name: &str) -> Self {
        DmVirtualMethod {
            fqn: fqn.to_string(),
            short_name: short_name.to_string(),
        }
    }

    pub fn fqn(&self) -> &str {
        self.fqn.as_str()
    }

    pub fn short_name(&self) -> &str {
        self.short_name.as_str()
    }
}

#[derive(Debug, Eq)]
pub struct DmImplementation {
    fqn: String,
    short_name: String,
    interface: Option<Rc<DmInterface>>,
    functions: Vec<Rc<DmFn>>,
    methods: Vec<Rc<DmMethod>>,
    fields: Vec<DmField>,
}

impl PartialEq for DmImplementation {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmImplementation {
    pub fn new(fqn: &str, short_name: &str, interface: Option<Rc<DmInterface>>) -> Self {
        DmImplementation {
            fqn: fqn.to_string(),
            short_name: short_name.to_string(),
            interface,
            functions: Vec::new(),
            methods: Vec::new(),
            fields: Vec::new(),
        }
    }

    pub fn fqn(&self) -> &str {
        self.fqn.as_str()
    }

    pub fn short_name(&self) -> &str {
        self.short_name.as_str()
    }

    pub fn interface(&self) -> Option<Rc<DmInterface>> {
        self.interface.clone()
    }

    pub fn functions(&self) -> &Vec<Rc<DmFn>> {
        &self.functions
    }

    pub fn add_function(&mut self, dm_fn: DmFn) {
        self.functions.push(Rc::new(dm_fn));
    }

    pub fn methods(&self) -> &Vec<Rc<DmMethod>> {
        &self.methods
    }

    pub fn add_method(&mut self, dm_method: DmMethod) {
        self.methods.push(Rc::new(dm_method));
    }

    pub fn find_method(&self, name: &str, self_object: &DvmObject) -> Option<&DmMethod> {
        for method in &self.methods {
            if method.dm_fn.fqn == name {
                return Some(method);
            } else if let Some(implements) = &method.implements {
                if implements.fqn == name {
                    return Some(method);
                }
            }
        }
        None
    }

    pub fn fields(&self) -> &Vec<DmField> {
        &self.fields
    }

    pub fn add_field(&mut self, dm_field: DmField) {
        self.fields.push(dm_field);
    }

    pub fn find_field(&self, name: &str, self_object: &DvmObject) -> Option<&DmField> {
        self.fields.iter().find(|field| field.name == name)
    }
}

#[derive(Debug, Eq)]
pub struct DmField {
    name: String,
    dm_type: DmType,
}

impl PartialEq for DmField {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}

impl DmField {
    pub fn new(name: &str, dm_type: DmType) -> Self {
        DmField {
            name: name.to_string(),
            dm_type,
        }
    }

    pub fn name(&self) -> &str {
        &self.name
    }

    pub fn dm_type(&self) -> &DmType {
        &self.dm_type
    }
}
